@@ -447,7 +447,7 @@ static int curl_retry_next_server(CURLM *curlm, CURL *curl, struct dload_payload
 		fseek(payload->localf, 0, SEEK_SET);
 	}
 
-	if(handle->dlcb && !payload->signature) {
+	if(handle->dlcb) {
 		alpm_download_event_retry_t cb_data;
 		cb_data.resume = payload->allow_resume;
 		handle->dlcb(handle->dlcb_ctx, payload->remote_name, ALPM_DOWNLOAD_RETRY, &cb_data);
@@ -613,23 +613,50 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 	/* Let's check if client requested downloading accompanion *.sig file */
 	if(!payload->signature && payload->download_signature && curlerr == CURLE_OK && payload->respcode < 400) {
 		struct dload_payload *sig = NULL;
+		char *url = payload->fileurl;
+		char *_effective_filename;
+		const char *effective_filename;
+		char *query;
+		const char *dbext = alpm_option_get_dbext(handle);
+		const char* realname = payload->destfile_name ? payload->destfile_name : payload->tempfile_name;
+		int len;
 
-		int len = strlen(effective_url) + 5;
+		STRDUP(_effective_filename, effective_url, GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
+		effective_filename = get_filename(_effective_filename);
+		query = strrchr(effective_filename, '?');
+
+		if(query) {
+			query[0] = '\0';
+		}
+
+		/* Only use the effective url for sig downloads if the effective_url contains .dbext or .pkg */
+		if(strstr(effective_filename, dbext) || strstr(effective_filename, ".pkg")) {
+			url = effective_url;
+		}
+
+		free(_effective_filename);
+
+		len = strlen(url) + 5;
 		CALLOC(sig, 1, sizeof(*sig), GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
 		MALLOC(sig->fileurl, len, FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
-		snprintf(sig->fileurl, len, "%s.sig", effective_url);
+		snprintf(sig->fileurl, len, "%s.sig", url);
 
 		if(payload->trust_remote_name) {
 			/* In this case server might provide a new name for the main payload.
 			 * Choose *.sig filename based on this new name.
 			 */
-			const char* realname = payload->destfile_name ? payload->destfile_name : payload->tempfile_name;
 			const char *final_file = get_filename(realname);
 			int remote_name_len = strlen(final_file) + 5;
 			MALLOC(sig->remote_name, remote_name_len, FREE(sig->fileurl); FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
 			snprintf(sig->remote_name, remote_name_len, "%s.sig", final_file);
 		}
 
+		/* force the filename to be realname + ".sig" */
+		int destfile_name_len = strlen(realname) + 5;
+		MALLOC(sig->destfile_name, destfile_name_len, FREE(sig->remote_name);
+				FREE(sig->fileurl); FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
+		snprintf(sig->destfile_name, destfile_name_len, "%s.sig", realname);
+
 		sig->signature = 1;
 		sig->handle = handle;
 		sig->force = payload->force;
@@ -761,15 +788,18 @@ static int curl_add_payload(alpm_handle_t *handle, CURLM *curlm,
 		GOTO_ERR(handle, ALPM_ERR_SERVER_BAD_URL, cleanup);
 	}
 
-	if(payload->remote_name && strlen(payload->remote_name) > 0) {
-		payload->destfile_name = get_fullpath(localpath, payload->remote_name, "");
+	if(!payload->random_partfile && payload->remote_name && strlen(payload->remote_name) > 0) {
+		if(!payload->destfile_name) {
+			payload->destfile_name = get_fullpath(localpath, payload->remote_name, "");
+		}
 		payload->tempfile_name = get_fullpath(localpath, payload->remote_name, ".part");
 		if(!payload->destfile_name || !payload->tempfile_name) {
 			goto cleanup;
 		}
 	} else {
-		/* URL doesn't contain a filename, so make a tempfile. We can't support
-		 * resuming this kind of download; partial transfers will be destroyed */
+		/* We want a random filename or the URL does not contain a filename, so download to a
+		 * temporary location. We can not support resuming this kind of download; any partial
+		 * transfers will be destroyed */
 		payload->unlink_on_fail = 1;
 
 		payload->localf = create_tempfile(payload, localpath);
@@ -817,6 +847,19 @@ cleanup:
 	return ret;
 }
 
+/*
+ * Use to sort payloads by max size in decending order (largest -> smallest)
+ */
+static int compare_dload_payload_sizes(const void *left_ptr, const void *right_ptr)
+{
+	struct dload_payload *left, *right;
+
+	left = (struct dload_payload *) left_ptr;
+	right = (struct dload_payload *) right_ptr;
+
+	return right->max_size - left->max_size;
+}
+
 /* Returns -1 if an error happened for a required file
  * Returns 0 if a payload was actually downloaded
  * Returns 1 if no files were downloaded and all errors were non-fatal
@@ -830,6 +873,10 @@ static int curl_download_internal(alpm_handle_t *handle,
 	int max_streams = handle->parallel_downloads;
 	int updated = 0; /* was a file actually updated */
 	CURLM *curlm = handle->curlm;
+	size_t payloads_size = alpm_list_count(payloads);
+
+	/* Sort payloads by package size */
+	payloads = alpm_list_msort(payloads, payloads_size, &compare_dload_payload_sizes);
 
 	while(active_downloads_num > 0 || payloads) {
 		CURLMcode mc;
@@ -978,11 +1025,20 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 			alpm_list_append(fetched, filepath);
 		} else {
 			struct dload_payload *payload = NULL;
+			char *c;
 
 			ASSERT(url, GOTO_ERR(handle, ALPM_ERR_WRONG_ARGS, err));
 			CALLOC(payload, 1, sizeof(*payload), GOTO_ERR(handle, ALPM_ERR_MEMORY, err));
 			STRDUP(payload->fileurl, url, FREE(payload); GOTO_ERR(handle, ALPM_ERR_MEMORY, err));
-			payload->allow_resume = 1;
+
+			c = strrchr(url, '/');
+			if(strstr(c, ".pkg")) {
+				/* we probably have a usable package filename to download to */
+				payload->allow_resume = 1;
+			} else {
+				payload->random_partfile = 1;
+			}
+
 			payload->handle = handle;
 			payload->trust_remote_name = 1;
 			payload->download_signature = (handle->siglevel & ALPM_SIG_PACKAGE);
